/*
    Wn: A Server for the HTTP
    File: wn/init.c
    Version 2.2.3
    
    Copyright (C) 1995-1999  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>

#ifndef NO_UNISTD_H
#include <unistd.h>
#endif

#ifndef NO_SYSLOG
#include <syslog.h>
#endif

#include <fcntl.h>
#include "wn.h"
#include "version.h"
#include "err.h"
#include "content.h"

#ifdef USE_VIRTUAL_HOSTS
extern char	*vhostlist[][4];
#endif

/* this may be needed if unistd.h isn't included... */
#ifndef SEEK_SET
#define SEEK_SET 0
#endif

#ifdef RFC931_TIMEOUT
extern void	get_rfc931();
#endif

#ifdef STANDALONE
#define PRINT_ERROR( x, y)	fprintf( stderr, x, y);
#define PRINT1_ERROR( x)	fprintf( stderr, x);
#else
#define PRINT_ERROR( x, y)	printf( x, y);
#define PRINT1_ERROR( x)	printf( x);
#endif

#ifdef NO_VOID_PTR
#define	void	int
#endif

#ifndef BSD_LIKE
extern long	timezone;
#endif

extern void	end_chunking();

extern char	*optarg;

extern int	optind;

#define		WN_COMMON	(0)
#define		WN_VERBOSE	(1)
#define		NCSA_LOG	(2)

#ifdef SET_UID
#ifndef USERID
#define USERID	(-2)
#define GROUPID	(-2)
#endif
#endif

#ifdef VERBOSELOG
static int	log_type = WN_VERBOSE;
#else
static int	log_type = WN_COMMON;
#endif

static void	dump_log();

char	rootdir[SMALLLEN],
	wn_tmpdir[sizeof(TEMPDIR) + TINYLEN] = {TEMPDIR},
	wn_empty[] = { "" },	/* safe place to park pointers */
	wnlogfile[SMALLLEN],
	errlogfile[SMALLLEN],
	pid_file[SMALLLEN],
	cfname[SMALLLEN],
	hostname[MAXHOSTNAMELEN],
	**mtypelist = NULL,
	**suflist = NULL;


int	admin_mode = FALSE,
	debug_log = FALSE,
 	user_id = USERID,
	group_id = GROUPID;


unsigned	serv_perm = 0,
		interface_num,
		acache_id,
		cache_id;

#define		LNOLOG	(0)
#define		LFILE	(1)
#define		LSYSLOG	(2)

static int	log_func = LNOLOG;

#ifndef NO_FLOCK
static void	locklog(),
		unlocklog();
#endif

#ifdef STANDALONE
static void	start_log(),
		restart();
#endif

static void	log_logfile();

static FILE	*logfp = NULL,
		*errlogfp = NULL;

#ifndef NO_SYSLOG
static void	log_syslog ();
#endif

static void	append_gmtoff();

void
wn_init( argc, argv)
int	argc;
char	*argv[];
{
	int	c,
		errflg = 0;


	port = DEFAULT_PORT;
	mystrncpy (wnlogfile, WN_LOGFILE, SMALLLEN);
	mystrncpy (errlogfile, WN_ERRLOGFILE, SMALLLEN);
	if ( *wnlogfile)
		log_func = LFILE;
	mystrncpy( rootdir, ROOT_DIR, SMALLLEN);

#ifdef SWN_PID_FILE
	mystrncpy( pid_file, SWN_PID_FILE, SMALLLEN);
#else
	*pid_file = '\0';
#endif

	mystrncpy( cfname, CACHEFNAME, SMALLLEN);
	mystrncpy( hostname, WN_HOSTNAME, MAXHOSTNAMELEN);


#ifdef FORBID_CGI
	serv_perm |= WN_FORBID_EXEC;
#endif

	while ((c = getopt(argc, argv, 
				"a:A:eEh:L:l:n:N:p:q:St:T:uv:V:dxz:")) != -1) {
		switch ((char) c) {
			case 'a':
				acache_id = (unsigned) atoi( optarg);
				serv_perm |= WN_ATRUSTED_UID;
				break;
			case 'A':
				acache_id = (unsigned) atoi( optarg);
				serv_perm |= WN_ATRUSTED_GID;
				break;
			case 'e':
				serv_perm |= WN_FORBID_EXEC;
				break;
			case 'E':
				serv_perm |= WN_RESTRICT_EXEC;
				break;
			case 'h':
				mystrncpy( hostname, optarg, MAXHOSTNAMELEN);
				break;
			case 'L':
				mystrncpy( wnlogfile, optarg, SMALLLEN);
				if ( *wnlogfile)
					log_func = LFILE;
				break;
			case 'l':
				mystrncpy( errlogfile, optarg, SMALLLEN);
				break;
#ifdef STANDALONE
			case 'n':
				user_id = (unsigned) atoi( optarg);
				break;
			case 'N':
				group_id = (unsigned) atoi( optarg);
				break;
			case 'p':
				port = atoi( optarg);
				break;
			case 'q':
				mystrncpy( pid_file, optarg, SMALLLEN);
				break;
#endif /* STANDALONE */

			case 'S':
#ifdef NO_SYSLOG
#ifdef STANDALONE
				fprintf( stderr,
					"Option -S unavailable, no syslog");
#endif /* STANDALONE */
				exit( 2);
#else
				wnlogfile[0] = '\0';
				errlogfile[0] = '\0';
				log_func = LSYSLOG;
#endif /* NO_SYSLOG */
				break;
			case 't':
				cache_id = (unsigned) atoi( optarg);
				serv_perm |= WN_TRUSTED_UID;
				break;
			case 'T':
				cache_id = (unsigned) atoi( optarg);
				serv_perm |= WN_TRUSTED_GID;
				break;
			case 'u':
				serv_perm |= WN_COMP_UID;
				break;
			case 'v':
				if ( strcasecmp( optarg, "common") == 0) {
					log_type = WN_COMMON;
					break;
				}
				if ( strcasecmp( optarg, "verbose") == 0) {
					log_type = WN_VERBOSE;
					break;
				}
				if ( strcasecmp( optarg, "ncsa") == 0) {
					log_type = NCSA_LOG;
					break;
				}
				PRINT_ERROR( "Unknown log type %.100s\n", optarg);
				break;
#ifdef USE_VIRTUAL_HOSTS
#ifdef VIRTUAL_HOSTS_FILE
			case 'V':
				mystrncpy( vhostfile, optarg, SMALLLEN);
				break;
#endif
#endif
			case 'd':
				/*should only be used for logging to file*/
				debug_log = TRUE;
				log_type = WN_VERBOSE;
				break;
			case 'x':
				admin_mode = TRUE;
				break;

			WN_EXTRA_OPTS

			case '?':
				errflg++;
		}
	}
	if (errflg) {
#ifdef STANDALONE
		PRINT1_ERROR("Usage: wn [-L logfile | -S] [-a uid |-A gid]\n");
		PRINT1_ERROR( "[-e | -E] [-p port] [-t uid | -T gid ] ");
		PRINT1_ERROR( "[-v log_type] [-V virtual host file]\n");
		PRINT1_ERROR( "[-u] [-q pid_file] [-h host] [topdir]\n");
#else
		PRINT1_ERROR( "Unknown option given to server");
#endif
		exit (2);
	}


#ifdef SET_UID
#ifndef STANDALONE  /* Otherwise this is done in standalone.c */
/*
 * Let's run as specified user and group if SET_UID defined.
 * This is only used with inetd and if inetd cannot change UID from root.
 * (i.e. Ultrix)
 */

 	setegid( (gid_t) group_id );
 	setgid(  (gid_t) group_id );   
 	seteuid( (uid_t) user_id );
 	setuid(  (uid_t) user_id );  /* This pulls rug out from under us */
#endif
#endif



	if ( *wnlogfile && !*errlogfile) {
		strcpy( errlogfile, wnlogfile);
	}

	if ( argv[optind] )
		mystrncpy( rootdir, argv[optind], SMALLLEN);


#ifndef NO_SYSLOG
	if (log_func == LSYSLOG)
	{

#ifdef LOG_DAEMON
		/* 4.3 style */
		openlog ("wn", LOG_PID | LOG_NDELAY, LOGFACILITY);
#else
		/* 4.2 style */
		openlog ("wn", LOG_PID);
#endif
	}
#endif
#ifndef STANDALONE
	open_wnlog( wnlogfile, errlogfile);
#endif




#ifdef USE_VIRTUAL_HOSTS
#ifdef VIRTUAL_HOSTS_FILE
	load_virtual();
#endif
#endif

#ifdef STANDALONE
	init_mime();
#endif

	WN_EXTRA_INIT
}

/*
 * void open_wnlog( logfile, errlog) is called once on startup to 
 * open the logfile, the error log file and to create a sub temp directory.
 * The point of the sub temp directory is to make temp files more 
 * secure.  If the server can't create this temp directory with 
 * mode 0700 or if there isn't already one there owned by the server
 * uid, then the server won't run for security reasons.  Unfortunately,
 * this tmp directory never gets removed when the server is shut down.
 */

void
open_wnlog( logfile, errlog)
char	*logfile,
	*errlog;
{

#ifndef FORBID_CGI
	char		buf[TINYLEN];
	struct stat	stat_buf;
	uid_t		my_id;

	my_id = getuid();
	Snprintf1( buf, TINYLEN, "/wn_tmp%d", my_id);
	mystrncat( wn_tmpdir, buf, sizeof(TEMPDIR) + TINYLEN);
	if ( lstat( wn_tmpdir, &stat_buf) == -1) {
		if ( (errno != ENOENT) || (mkdir( wn_tmpdir, 0700) != 0) ) {
			PRINT_ERROR( err_m[122], wn_tmpdir);
			exit( 2);
		}
	}
	else if ( (!S_ISDIR(stat_buf.st_mode))
				|| (stat_buf.st_uid != my_id)
				|| (chmod( wn_tmpdir, 0700) != 0) ) {
		PRINT_ERROR( err_m[122], wn_tmpdir);
		PRINT_ERROR( "errno=%d\n", errno);
		exit( 2);
	}
#endif

	if ( *logfile != '\0')
		if ( (logfp = fopen( logfile, "a")) == NULL ) {
			PRINT_ERROR( err_m[0], logfile);
			PRINT_ERROR( "errno=%d\n", errno);
			exit( 2);
		}
	if ( *errlog != '\0') {
		if ( streq( errlog, logfile))
			errlogfp = logfp;

		else if ( (errlogfp = fopen( errlog, "a")) == NULL ) {
			PRINT_ERROR( err_m[0], errlog);
			PRINT_ERROR( "errno=%d\n", errno);
			exit( 2);
		}
	}
#ifdef STANDALONE
	start_log( FALSE);
	signal( SIGHUP, restart);
#endif
}


#ifdef STANDALONE
static void
restart( )
{
	signal( SIGHUP, SIG_IGN);

#ifdef USE_VIRTUAL_HOSTS
#ifdef VIRTUAL_HOSTS_FILE
	load_virtual();
#endif
#endif
	if ( *wnlogfile) {
		if ( logfp) 
			fclose( logfp);
		if ( (logfp = fopen( wnlogfile, "a")) == NULL ) {
			PRINT_ERROR( err_m[0], wnlogfile);
			PRINT_ERROR( "restart errno=%d\n", errno);
			exit( 2);
		}
	}

	if ( *errlogfile) {
		if ( streq( errlogfile, wnlogfile))
			errlogfp = logfp;
		else {
			if ( errlogfp)
				fclose( errlogfp);
			if ( (errlogfp = fopen( errlogfile, "a")) == NULL ){
				PRINT_ERROR( err_m[0], errlogfile);
				PRINT_ERROR( "restart errno=%d\n", errno);
				exit( 2);
			}
		}
	}

	start_log( TRUE);
	init_mime();
	signal( SIGHUP, restart);
}
#endif /* STANDALONE */


void
logerr(  msg, msg2)
char	*msg,
	*msg2;
{

#ifndef NO_FLOCK
	struct flock	lck;
#endif

	time_t	clock;
	struct tm *ltm;

	char	status[TINYLEN],
		date[TINYLEN],
		xmsg2[MIDLEN + 1];


	if ( msg2 == NULL)
		xmsg2[0] = '\0';
	else {
		strncpy( xmsg2, msg2, SMALLLEN);
		/* can't use mystrncpy() since it calls us */
		xmsg2[MIDLEN] = '\0';
	}

	if ((log_func == LNOLOG) && !*errlogfile)
		return;

	get_remote_info( );

#ifndef NO_SYSLOG
	if ( (log_func == LSYSLOG) && !*errlogfile) {
		log_syslog( ERRLOG_PRIORITY, this_rp, msg, xmsg2);
		return;
	}
#endif

	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( date, TINYLEN, "%d/%h/%Y:%T", ltm);
	append_gmtoff( date, ltm);

#ifndef NO_FLOCK
	locklog( &lck, fileno( errlogfp) );
#endif

	if ( outheadp && (*outheadp->status)) {
		strncpy( status, outheadp->status, 4);
		status[4] = '\0';
	}
	else
		strcpy( status, "500");

	fseek( errlogfp, 0L, 2);
	fprintf( errlogfp, "%.200s - - [%.32s] \"%.300s\" %.32s -",
		 this_conp->remotehost,	date, this_rp->request, status);

	if ( log_type == WN_VERBOSE ) {
		fprintf( errlogfp, " <(%d/%d) %.200s: %.200s>",
			this_conp->pid, this_conp->trans_cnt, msg, xmsg2);
	}
	fprintf( errlogfp, "\n");
	(void) fflush( errlogfp);


#ifndef NO_FLOCK
	unlocklog( &lck, fileno( errlogfp) );   
#endif

}

void
daemon_logerr(  msg, msg2, error)
char	*msg,
	*msg2;
int	error;
{

	FILE		*errfp;
#ifndef NO_FLOCK
	struct flock	lck;
#endif
	time_t	clock;
	struct tm *ltm;

	char	buf[SMALLLEN],
		date[TINYLEN];


	if ((log_func == LNOLOG) && !*errlogfile)
		return;


#ifndef NO_SYSLOG
	if ( (log_func == LSYSLOG) && !*errlogfile) {
		Snprintf3( buf, SMALLLEN,
			"none - - [] \"none\" 500 0  <%.100s: %.100s errno=%d> ",
				msg, msg2, error);

		if ( log_type == WN_VERBOSE ) {
#ifdef USE_VIRTUAL_HOSTS
			syslog( ERRLOG_PRIORITY, "<> <> <> <0>\n");
#else
			syslog( ERRLOG_PRIORITY, "<> <> <>\n");
#endif
		}
		else
			mystrncat( buf, "\n", SMALLLEN);

		syslog( ERRLOG_PRIORITY, buf);
		return;
	}
#endif /* NO_SYSLOG */

	if ( errlogfp)
		errfp = errlogfp;
	else
		errfp = stderr;

	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( date, TINYLEN, "%d/%h/%Y:%T", ltm);
	append_gmtoff( date, ltm);

#ifndef NO_FLOCK
	if ( errlogfp)
		locklog( &lck, fileno( errlogfp) );
#endif

	fseek( errfp, 0L, 2);   
	fprintf(errfp, "none - - [%.32s] \"none\" 500 0",  date);
	if ( log_type == WN_VERBOSE ) {
		fprintf( errfp, " <%.200s %.200s: errno=%d> <> <> <>",
					msg, msg2, error);
#ifdef USE_VIRTUAL_HOSTS
		fprintf( errfp, " <0>");
#endif
	}

	fprintf( errfp, "\n");
	(void) fflush( errfp);

#ifndef NO_FLOCK
	if ( errlogfp)
		unlocklog( &lck, fileno( errlogfp) );   
#endif
}


void
writelog( ip, msg, msg2)
Request	*ip;
char	*msg,
	*msg2;
{
	if (log_func == LFILE)
		log_logfile( ip, msg, msg2);
#ifndef NO_SYSLOG
	else if ( log_func == LSYSLOG)
		log_syslog( LOG_PRIORITY, ip, msg, msg2);
#endif
}


/* Write debug messages into the log file. */

void
write_debug(n, msg, msg2)
int	n; /*ignored for now, could become the debug level later*/
char	*msg, *msg2;
{
	if (log_func == LFILE) {
		fprintf(logfp, "%.300s %.300s\n", msg, msg2);
		(void) fflush( logfp);
	}
#ifndef NO_SYSLOG
	else if ( log_func == LSYSLOG) {
		if ( strlen( msg) > 128 )
			msg[127] = '\0';
		if ( strlen( msg2) > 128 )
			msg2[127] = '\0';
		log_syslog( LOG_PRIORITY, this_rp, msg, msg2);
	}
#endif
}


static void
log_logfile( ip, msg, msg2)
Request	*ip;
char	*msg,
	*msg2;
{
	time_t	clock;
	struct tm *ltm;

	register char	*cp,
			*cp2;

	char	*authname,
		bytes[TINYLEN],
		status[TINYLEN],
		date[TINYLEN],
		lbuf[2*SMALLLEN];

	if ( outheadp && (*outheadp->status)) {
		cp2 = status;
		cp = outheadp->status;
		while ( *cp && !isspace(*cp) && ((cp - outheadp->status) < 10))
			*cp2++ = *cp++;
		*cp2 = '\0';
	}
	else
		strcpy( status, "200");

	Snprintf1( bytes, TINYLEN, "%ld", ip->logcount);

	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( date, TINYLEN, "%d/%h/%Y:%T", ltm);
	append_gmtoff( date, ltm);

	cp = cp2 = this_conp->log_ptr;

	authname = ( *this_rp->authuser ? this_rp->authuser : "-");
	Snprintf5( lbuf, 2*SMALLLEN, "%.100s [%.32s] \"%.200s\" %.32s %.32s",
			   authname, date, ip->request, status, bytes);
	cp2 += strlen( lbuf);
	if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
		dump_log();
		cp = cp2 = this_conp->logbuf;
	}
	strcpy( cp , lbuf);
	cp = cp2;

	if ( log_type == WN_VERBOSE ) {
		if ( streq( status, "206")) {
			char *rangep;

			rangep = outheadp->range;
			if ( !*rangep)
				rangep = "multipart";

			Snprintf5( lbuf, 2*SMALLLEN, 
				" <(%d/%d) %.100s (%.100s): %.200s>",
				this_conp->pid, this_conp->trans_cnt, 
				log_m[23], rangep, msg2);
		}
		else {
			Snprintf4( lbuf, 2*SMALLLEN, " <(%d/%d) %.200s: %.200s>",
			this_conp->pid, this_conp->trans_cnt, msg, msg2);
		}
		cp2 += strlen( lbuf);
		if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
			dump_log();
			cp = cp2 = this_conp->logbuf;
		}

 		
		strcpy( cp, lbuf);
		cp = cp2;

		Snprintf1( lbuf, 2*SMALLLEN," <%.500s>", inheadp->ua);
		cp2 += strlen( lbuf);
		if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
			dump_log();
			cp = cp2 = this_conp->logbuf;
		}

		strcpy( cp , lbuf);
		cp = cp2;

		Snprintf1( lbuf, 2*SMALLLEN, " <%.500s>", inheadp->referrer);
		cp2 += strlen( lbuf);
		if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
			dump_log();
			cp = cp2 = this_conp->logbuf;
		}
		strcpy( cp , lbuf);
		cp = cp2;

		Snprintf1( lbuf, 2*SMALLLEN, " <%.500s>", inheadp->cookie);
		cp2 += strlen( lbuf);
		if ( cp2 >= this_conp->logbuf + (LOGBUFLEN)) {
			dump_log();
			cp = cp2 = this_conp->logbuf;
		}
		strcpy( cp , lbuf);
		cp = cp2;

		Snprintf1( lbuf, 2*SMALLLEN, " <%.500s>", inheadp->xforwardedfor);
		cp2 += strlen( lbuf);
		if ( cp2 >= this_conp->logbuf + (LOGBUFLEN - 64)) {
			dump_log();
			cp = cp2 = this_conp->logbuf;
		}
		strcpy( cp , lbuf);
		cp = cp2;

#ifdef USE_VIRTUAL_HOSTS
		if ( (interface_num > 0) && *vhostlist[interface_num -1][3] )
			Snprintf1( cp, TINYLEN, " <%.32s>", 
					vhostlist[interface_num - 1][3] );
		else
			Snprintf1( cp, TINYLEN, " <%d>", interface_num);
		cp += strlen( cp);
#endif
	}
	else if ( log_type == NCSA_LOG) {
		Snprintf1( lbuf, 2*SMALLLEN, " \"%.200s\"", inheadp->referrer);
		cp2 += strlen( lbuf);
		if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
			dump_log();
			cp = cp2 = this_conp->logbuf;
		}
		strcpy( cp , lbuf);
		cp = cp2;

		Snprintf1( lbuf, SMALLLEN, " \"%.100s\"", inheadp->ua);
		cp2 += strlen( lbuf);
		if ( cp2 >= this_conp->logbuf + LOGBUFLEN) {
			dump_log();
			cp = cp2 = this_conp->logbuf;
		}
		strcpy( cp , lbuf);
		cp = cp2;
	}

	*cp++ =  '\n';
	*cp =  '\0';
	this_conp->log_ptr = cp;
}


#ifndef NO_FLOCK
static void
locklog(lck, fd)
struct flock	*lck;
int	fd;
{
	lck->l_type = F_WRLCK;
	lck->l_whence = SEEK_SET;
	lck->l_start = 0L;
	lck->l_len = 0L;
	fcntl(fd, F_SETLKW, lck);
}

static void
unlocklog( lck, fd)
struct flock	*lck;
int	fd;
{
	lck->l_type = F_UNLCK;
	fcntl( fd, F_SETLKW, lck);
}
#endif


#ifndef NO_SYSLOG

static void
log_syslog( priority, ip, msg, msg2)
int	priority;
Request	*ip;
char	*msg,
	*msg2;
{
	char	*authname,
		*rfc931p,
		bytes[TINYLEN],
		status[TINYLEN];

	if ( *outheadp->status) {
		strncpy( status, outheadp->status, 4);
		/* can't use mystrncpy as it may call us */
		status[3] = '\0';
	}
	else
		strcpy( status, "200");

	Snprintf1( bytes, TINYLEN, "%ld", ip->logcount);

	rfc931p = ( *this_conp->rfc931name ? this_conp->rfc931name : "-");
	authname = ( *ip->authuser ? ip->authuser : "-");

	get_remote_info();

	syslog( priority,
		"%.100s %.32s %.32s [] \"%.100s\" %.10s %.10s  <%.100s: %.100s>\n", 
		this_conp->remotehost, rfc931p, authname, ip->request, status, 
		bytes, msg, msg2);

	if ( log_type == WN_VERBOSE ) {
#ifdef USE_VIRTUAL_HOSTS
		char			label[SMALLLEN];

		if ( (interface_num > 0 ) && *vhostlist[interface_num - 1][3] )
			Snprintf1( label, SMALLLEN, " <%.64s>", 
					vhostlist[interface_num - 1][3]);
		else
			Snprintf1( label, TINYLEN, " <%d>", interface_num );

		syslog( priority, "<%.100s> <%.100s> <%.100s> <%.100s>\n",
			inheadp->ua, inheadp->referrer, inheadp->cookie, label);
#else
		syslog( priority, "<%.100s> <%.100s> <%.100s>\n",
			inheadp->ua, inheadp->referrer, inheadp->cookie);
#endif
	}
}
#endif /* ifndef NO_SYSLOG */


static void
append_gmtoff( date, ltm)
char	*date;
struct tm *ltm;
{
	register char	*cp;
	long		tz;
	char 		sign;

#ifdef BSD_LIKE
	tz = ltm->tm_gmtoff;
#else
	tz = - timezone;
	if( ltm->tm_isdst)
		tz += 3600;
#endif
	sign = ( tz > 0 ? '+' : '-');
	tz = ( tz > 0 ? tz :  -tz);
	cp = date;
	while ( *cp)
		cp++;
	
	Snprintf3( cp, 7, " %c%02ld%02ld", sign, tz/3600, tz % 3600);
}

#ifdef STANDALONE
static void
start_log( restart)
int	restart;
{
	time_t	clock;
	struct tm *ltm;
	char	*cp,
		startdate[TINYLEN];

	if ( log_func != LFILE)
		return;
	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( startdate, TINYLEN, "%d/%h/%Y:%T", ltm);
	cp = ( restart ? "Restarting" : "Starting");
	fprintf(logfp, "\n%.32s: %s %.32s at port %d with pid %d\n", 
			startdate, cp, VERSION, port, getpid() );

	(void) fflush(logfp);
}
#endif


void
wn_abort( )
{

	this_conp->keepalive = FALSE;
	this_rp->type = RTYPE_FINISHED;
	this_rp->status |= (WN_ABORTED + WN_ERROR);
	
}

void
wn_exit( status)
int	status;
{
	int	n;
	char	buf[MIDLEN];

	if ( this_conp->chunk_status & WN_USE_CHUNK)
		end_chunking();
	flush_outbuf();
	if ( *inheadp->tmpfile) {
		unlink( inheadp->tmpfile);
		*inheadp->tmpfile = '\0';
	}

	shutdown( fileno( stdin), 1);
	dump_log();

	if ( status != 0) {
		/* error condition */
		signal( SIGALRM, SIG_DFL);
		alarm( 3);  	
		/* Give ourselves at most 3 seconds to shutdown */
		while ( (n = WN_read( (fileno( stdin)), buf, MIDLEN)) > 0 ) {
			; /* Collect stuff from client and throw it away */
		}
	}

	close( fileno( stdin));
	exit( status);
}


static void
dump_log()
{
	register char	*cp,
			*cp2;

	char		*rfc931p,
			logline[MIDLEN];

#ifndef NO_FLOCK
	struct flock	lck;
#endif

	if ( this_conp->logbuf == this_conp->log_ptr)
		return;

	*(this_conp->log_ptr) = '\0';

	get_remote_info( );


#ifndef NO_FLOCK
	locklog( &lck, fileno( logfp) );
#endif

#ifdef RFC931_TIMEOUT
	get_rfc931();
#endif
	rfc931p = ( *this_conp->rfc931name ? this_conp->rfc931name : "-");
	fseek( logfp, 0L, 2);

	cp = this_conp->logbuf;
	cp2 = logline;

	while ( *cp && ( cp2 < (logline + MIDLEN))) {
		*cp2 = *cp;
		if ( *cp2 == '\n') {
			*++cp2 = '\0';
			fprintf( logfp, "%.255s %.100s %.1000s", 
				this_conp->remotehost, rfc931p, logline);
			cp2 = logline;
			cp++;
			continue;
		}
		cp2++;
		cp++;
	}
	(void) fflush( logfp);
	this_conp->log_ptr = this_conp->logbuf;

#ifndef NO_FLOCK
	unlocklog( &lck, fileno( logfp) );   
#endif

}
void
get_mtype( suffix )
char	*suffix;
{
	int	i,
		unknown_mime = TRUE;

	for ( i = 0; mimelist[i][0] != NULL; i++) {
		if ( streq( mimelist[i][0], suffix)) {
			mystrncpy( this_rp->contype, mimelist[i][1], SMALLLEN);
			this_rp->content_type = this_rp->contype;
			unknown_mime = FALSE;
			break;
		}
	}

#ifdef STANDALONE
	if ( unknown_mime && (suflist != NULL)) {
		for ( i = 0; suflist[i] != NULL; i++) {
			if ( streq( suflist[i], suffix)) {
				mystrncpy( this_rp->contype, 
						mtypelist[i], SMALLLEN);
				this_rp->content_type = this_rp->contype;
				break;
			}
		}
	}
#endif  /* STANDALONE */
}

#ifdef STANDALONE
int
needsuffix( suf)
char	*suf;

{
	int	i;

	for ( i = 0; mimelist[i][0] != NULL; i++) {
		if ( streq( suf, mimelist[i][0]))
			return FALSE;
	}
	return TRUE;
}
#endif  /* STANDALONE */
